from django.core.management.base import BaseCommand, CommandError

from micro_framework.jwt_auth.redis_token.token_manager import RedisAccessTokenManager


class Command(BaseCommand):
    help = 'Revoke token store in Redis'

    def add_arguments(self, parser):
        parser.add_argument('-a', '--all', action='store_true', help='Revoke all token storing in redis')
        parser.add_argument('-uid', '--user_id', type=str, help='Revoke all token belong to user_id')

    def handle(self, *args, **kwargs):
        redis_access_tokens = RedisAccessTokenManager()
        all = kwargs['all']
        if all:
            redis_access_tokens.delete_all()
            print("All access token has been revoke")

        user_id = kwargs.get('user_id', None)
        if user_id:
            list_user_token = redis_access_tokens.list(user_id=user_id)
            for redis_token in list_user_token:
                redis_token.delete()
            msg = "All access token belong to {} has been revoke".format(user_id)
            print(msg)



from django.utils.translation import ugettext_lazy as _
from rest_framework import HTTP_HEADER_ENCODING, authentication

from micro_framework.jwt_auth.exceptions import AuthenticationFailed, InvalidToken, TokenError
from micro_framework.jwt_auth.models import TokenUser
from micro_framework.settings import api_settings
from micro_framework.jwt_auth.state import User

AUTH_HEADER_TYPES = api_settings.AUTH_HEADER_TYPES

if not isinstance(api_settings.AUTH_HEADER_TYPES, (list, tuple)):
    AUTH_HEADER_TYPES = (AUTH_HEADER_TYPES,)

AUTH_HEADER_TYPE_BYTES = set(
    h.encode(HTTP_HEADER_ENCODING)
    for h in AUTH_HEADER_TYPES
)


class JWTAuthentication(authentication.BaseAuthentication):
    """
    An authentication plugin that authenticates requests through a JSON web
    token provided in a request header.
    """
    www_authenticate_realm = 'api'

    def authenticate(self, request):
        header = self.get_header(request)
        if header is None:
            return None

        raw_token = self.get_raw_token(header)
        if raw_token is None:
            return None

        validated_token = self.get_validated_token(raw_token)

        return self.get_user(validated_token), validated_token

    def authenticate_header(self, request):
        return '{0} realm="{1}"'.format(
            AUTH_HEADER_TYPES[0],
            self.www_authenticate_realm,
        )

    def get_header(self, request):
        """
        Extracts the header containing the JSON web token from the given
        request.
        """
        header = request.META.get('HTTP_AUTHORIZATION')

        if isinstance(header, str):
            # Work around django test client oddness
            header = header.encode(HTTP_HEADER_ENCODING)

        return header

    def get_raw_token(self, header):
        """
        Extracts an unvalidated JSON web token from the given "Authorization"
        header value.
        """
        parts = header.split()

        if len(parts) == 0:
            # Empty AUTHORIZATION header sent
            return None

        if parts[0] not in AUTH_HEADER_TYPE_BYTES:
            # Assume the header does not contain a JSON web token
            return None

        if len(parts) != 2:
            raise AuthenticationFailed(
                _('Authorization header must contain two space-delimited values'),
                code='bad_authorization_header',
            )

        return parts[1]

    def get_validated_token(self, raw_token):
        """
        Validates an encoded JSON web token and returns a validated token
        wrapper object.
        """
        messages = []
        for AuthToken in api_settings.AUTH_TOKEN_CLASSES:
            try:
                return AuthToken(raw_token)
            except TokenError as e:
                messages.append({'token_class': AuthToken.__name__,
                                 'token_type': AuthToken.token_type,
                                 'message': e.args[0]})

        raise InvalidToken({
            'detail': _('Given token not valid for any token type'),
            'messages': messages,
        })

    def get_user(self, validated_token):
        """
        Attempts to find and return a user using the given validated token.
        """
        try:
            user_id = validated_token[api_settings.USER_ID_CLAIM]
        except KeyError:
            raise InvalidToken(_('Token contained no recognizable user identification'))

        try:
            user = User.objects.get(**{api_settings.USER_ID_FIELD: user_id})
        except User.DoesNotExist:
            raise AuthenticationFailed(_('User not found'), code='user_not_found')

        if not user.is_active:
            raise AuthenticationFailed(_('User is inactive'), code='user_inactive')

        return user


class JWTTokenUserAuthentication(JWTAuthentication):
    def get_user(self, validated_token):
        """
        Returns a stateless user object which is backed by the given validated
        token.
        """
        if api_settings.USER_ID_CLAIM not in validated_token:
            # The TokenUser class assumes tokens will have a recognizable user
            # identifier claim.
            raise InvalidToken(_('Token contained no recognizable user identification'))

        return TokenUser(validated_token)



# Create your models here.
from django.contrib.auth import models as auth_models
from django.db import models
from django.db.models.manager import EmptyManager
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from micro_framework.jwt_auth.compat import CallableFalse, CallableTrue
from micro_framework.settings import api_settings


class TokenUser:
    """
    A dummy user class modeled after django.contrib.auth.models.AnonymousUser.
    Used in conjunction with the `JWTTokenUserAuthentication` backend to
    implement single sign-on functionality across services which share the same
    secret key.  `JWTTokenUserAuthentication` will return an instance of this
    class instead of a `User` model instance.  Instances of this class act as
    stateless user objects which are backed by validated tokens.
    """
    # User is always active since Simple JWT will never issue a token for an
    # inactive user
    is_active = True
    default_fields = ['username', 'is_staff', 'is_superuser', 'roles',
                      'id', 'pk']

    _groups = EmptyManager(auth_models.Group)
    _user_permissions = EmptyManager(auth_models.Permission)

    def __init__(self, token):
        self.token = token

        for key, value in self.token.payload.items():
            if key not in self.default_fields:
                setattr(self, key, value)

    def __str__(self):
        return 'TokenUser {}'.format(self.id)

    @cached_property
    def id(self):
        return self.token[api_settings.USER_ID_CLAIM]

    @cached_property
    def pk(self):
        return self.id

    @cached_property
    def username(self):
        return self.token.get('username', '')
    
    @cached_property
    def email(self):
        return self.token.get('email', '')

    @cached_property
    def is_staff(self):
        return self.token.get('is_staff', False)

    @cached_property
    def is_superuser(self):
        return self.token.get('is_superuser', False)
    
    @property
    def roles_list(self):
        return self.token.get('roles', [])

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)

    def save(self):
        raise NotImplementedError('Token users have no DB representation')

    def delete(self):
        raise NotImplementedError('Token users have no DB representation')

    def set_password(self, raw_password):
        raise NotImplementedError('Token users have no DB representation')

    def check_password(self, raw_password):
        raise NotImplementedError('Token users have no DB representation')

    @property
    def groups(self):
        return self._groups

    @property
    def user_permissions(self):
        return self._user_permissions

    def get_group_permissions(self, obj=None):
        return set()

    def get_all_permissions(self, obj=None):
        return set()

    def has_perm(self, perm, obj=None):
        return False

    def has_perms(self, perm_list, obj=None):
        return False

    def has_module_perms(self, module):
        return False

    @property
    def is_anonymous(self):
        return CallableFalse

    @property
    def is_authenticated(self):
        return CallableTrue

    def get_username(self):
        return self.username

class Policy(models.Model):
    name = models.CharField(max_length=255, unique=True)
    rules = models.TextField()
    notes = models.TextField(blank=True)

    class Meta:
        db_table = 'policies'
        verbose_name = _("Policy")
        verbose_name_plural = _("Policies")

    def __str__(self):
        return self.name



from django.core.management.base import BaseCommand, CommandError

from micro_framework.jwt_auth.models import Policy


class Command(BaseCommand):
    help = 'Create Policy for service'

    def create_service_policy(self, policy_name, policy_rule):
        policy = Policy.objects.get_or_create(name=policy_name, rules=policy_rule)
        return policy[0]

    def add_arguments(self, parser):
        parser.add_argument('policyname', type=str)
        parser.add_argument('policyrule', type=str)

    def handle(self, *args, **kwargs):
        policy_name = kwargs['policyname']
        policy_rule = kwargs['policyrule']

        try:
            token = self.create_service_policy(policy_name, policy_rule)
        except Exception as ex:
            raise CommandError(
                'Cannot create the Policy {} '.format(
                    policy_name)
            )
        self.stdout.write(
            'Created Policy ({}) with rule ({})'.format(policy_name, policy_rule))



"""
MIXINS
"""
from django.utils.translation import ugettext_lazy as _

from micro_framework.jwt_auth import acls
from micro_framework.settings import api_settings
from micro_framework.helpers import check_client_ip_and_token_ip

class PermissionRequiredMixin:

    permission = None

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        ## Checking permission request
        self.check_require_permissions(request)
        self.check_validate_ip_source(request)

    def check_require_permissions(self, request):
        """
        Checking permission
        """
        if self.permission is not None:
            if not acls.action_allowed(request, self.permission):
                self.permission_denied(
                    request, message=_('You do not have permission to perform this action.')
                )

    def check_validate_ip_source(self, request):
        """
        Checking real_ip in request with real_ip in payload token
        """
        if api_settings.VALIDATE_SOURCE_IP and not check_client_ip_and_token_ip(request):
            self.permission_denied(
                request, message=_('Token is invalid.')
            )



from calendar import timegm
from datetime import datetime

from django.conf import settings
from django.utils.functional import lazy
from django.utils.timezone import is_naive, make_aware, utc

class AclPermission(object):
    def __init__(self, app, action):
        self.app = app
        self.action = action

    @classmethod
    def as_string(cls, string_rule):
        data = str(string_rule).split(":")
        app = data[0]
        action = data[1]
        return cls(app=app, action=action)

    @property
    def string_rule(self):
        return self.app + ':' + self.action

def make_utc(dt):
    if settings.USE_TZ and is_naive(dt):
        return make_aware(dt, timezone=utc)

    return dt


def aware_utcnow():
    return make_utc(datetime.utcnow())


def datetime_to_epoch(dt):
    return timegm(dt.utctimetuple())


def datetime_from_epoch(ts):
    return make_utc(datetime.utcfromtimestamp(ts))


def format_lazy(s, *args, **kwargs):
    return s.format(*args, **kwargs)

format_lazy = lazy(format_lazy, str)



import redis

from micro_framework.settings import api_settings

redis_backend = None
if api_settings.REDIS_ENABLE:
    redis_backend = redis.Redis(host=api_settings.REDIS_HOST,
                                port=api_settings.REDIS_PORT,
                                db=api_settings.REDIS_DB,
                                password=api_settings.REDIS_PASSWORD)



from django.utils.translation import ugettext_lazy as _

from micro_framework.jwt_auth.exceptions import InvalidToken
from micro_framework.jwt_auth.models import Policy
from micro_framework.settings import api_settings


def match_rules(rules, app, action):
    """
    This will match rules found in Policy.
    """
    for rule in rules.split(','):
        try:
            rule_app, rule_action = rule.split(':')

            # Loại bỏ dấu cách nếu thừa
            rule_app = rule_app.replace(" ", "")
            rule_action = rule_action.replace(" ", "")

            if rule_app == '*' or rule_app == app:
                if rule_action == '*' or rule_action == action or action == '%':
                    return True
        except Exception as ex:
            continue
    return False

def action_allowed(request, permission):
    return action_allowed_user(request.user, permission)

def action_allowed_user(user, permission):
    if not user.is_authenticated:
        return False

    return any(
        match_rules(policy.rules, permission.app, permission.action)
        for policy in roles_to_policies(user)
    )

def roles_to_policies(user):
    list_role = []
    for role_user in user.roles_list:
        service_name, role_name = role_user.split(':')
        if service_name == api_settings.SERVICE_NAME:
            list_role.append(role_name)
    list_policy = []
    for role_name in list_role:
        try:
            user_policy = Policy.objects.get(name=role_name)
            list_policy.append(user_policy)
        except Exception as ex:
            messages = []
            messages.append({'message': str(ex)})
            raise InvalidToken({
                'detail': _('Given token not valid for any token type'),
                'messages': messages,
            })
    return list_policy



import jwt
from django.utils.translation import ugettext_lazy as _
from jwt import InvalidTokenError

from micro_framework.jwt_auth.exceptions import TokenBackendError
from micro_framework.jwt_auth.utils import format_lazy

ALLOWED_ALGORITHMS = (
    'HS256',
    'HS384',
    'HS512',
    'RS256',
    'RS384',
    'RS512',
)


class TokenBackend:
    def __init__(self, algorithm, signing_key=None, verifying_key=None, audience=None, issuer=None):
        if algorithm not in ALLOWED_ALGORITHMS:
            raise TokenBackendError(format_lazy(_("Unrecognized algorithm type '{}'"), algorithm))

        self.algorithm = algorithm
        self.signing_key = signing_key
        self.audience = audience
        self.issuer = issuer
        if algorithm.startswith('HS'):
            self.verifying_key = signing_key
        else:
            self.verifying_key = verifying_key

    def encode(self, payload):
        """
        Returns an encoded token for the given payload dictionary.
        """
        jwt_payload = payload.copy()
        if self.audience is not None:
            jwt_payload['aud'] = self.audience
        if self.issuer is not None:
            jwt_payload['iss'] = self.issuer

        token = jwt.encode(jwt_payload, self.signing_key, algorithm=self.algorithm)
        return token.decode('utf-8')

    def decode(self, token, verify=True):
        """
        Performs a validation of the given token and returns its payload
        dictionary.

        Raises a `TokenBackendError` if the token is malformed, if its
        signature check fails, or if its 'exp' claim indicates it has expired.
        """
        try:
            return jwt.decode(token, self.verifying_key, algorithms=[self.algorithm], verify=verify,
                              audience=self.audience, issuer=self.issuer,
                              options={'verify_aud': self.audience is not None})
        except InvalidTokenError:
            raise TokenBackendError(_('Token is invalid or expired'))



import redis

from micro_framework.settings import api_settings


def redis_client(host=None, port=None, db=None, password=None):
    """
    Create Client connected to Redis Server
    """
    if not host:
        host = api_settings.REDIS_HOST
    if not port:
        port = api_settings.REDIS_PORT
    if not db:
        db = api_settings.REDIS_DB
    if not password:
        password = api_settings.REDIS_PASSWORD
    return redis.Redis(host=host, port=port, db=db, password=password)
